local reg = debug.getregistry()
local lp = love.physics

local newBody = lp.newBody
function reg.World:newBody(x, y, t)
  return newBody(self, x, y, t)
end

local destroyBody = reg.Body.destroy
function reg.World:destroyBody(b)
  destroyBody(b)
end

local destroyJoint = reg.Joint.destroy
function reg.World:destroyJoint(j)
  destroyJoint(j)
end

local newPolygonShape = lp.newPolygonShape
local newFixture = lp.newFixture
function reg.Body:newPolygon(...)
  local s = newPolygonShape(...)
  return newFixture(self, s)
end

local newRectangleShape = lp.newRectangleShape
function reg.Body:newBox(w, h, lx, ly, la)
  lx, ly = lx or 0, ly or 0
  la = la or 0
  local s = newRectangleShape(lx, ly, w*2, h*2, la)
  return newFixture(self, s)
end

local newCircleShape = lp.newCircleShape
function reg.Body:newCircle(radius, lx, ly)
  lx, ly = lx or 0, ly or 0
  local s = newCircleShape(lx, ly, radius)
  return newFixture(self, s)
end

local newChainShape = lp.newChainShape
function reg.Body:newChain(vertices, loop)
  local s = newChainShape(loop, vertices)
  return newFixture(self, s)
end

local newEdgeShape = lp.newEdgeShape
function reg.Body:newEdge(x1, y1, x2, y2)
  local s = newEdgeShape(x1, y1, x2, y2)
  return newFixture(self, s)
end

local newRevoluteJoint = lp.newRevoluteJoint
function reg.World:newRevoluteJoint(a, b, x, y, cc)
  return newRevoluteJoint(a, b, x, y, cc)
end

local newPrismaticJoint = lp.newPrismaticJoint
function reg.World:newPrismaticJoint(a, b, x, y, ax, ay, cc)
  return newPrismaticJoint(a, b, x, y, ax, ay, cc)
end

local newDistanceJoint = lp.newDistanceJoint
function reg.World:newDistanceJoint(a, b, p1x, p1y, p2x, p2y, cc)
  return newDistanceJoint(a, b, p1x, p1y, p2x, p2y, cc)
end

local newRopeJoint = lp.newRopeJoint
local sqrt = math.sqrt
function reg.World:newRopeJoint(a, b, p1x, p1y, p2x, p2y, l, cc)
  if l == nil then
    local lx, ly = p1x - p2x, p1y - p2y
    l = sqrt(lx*lx + ly*ly)
  end
  return newRopeJoint(a, b, p1x, p1y, p2x, p2y, l, cc)
end

local newPulleyJoint = lp.newPulleyJoint
function reg.World:newPulleyJoint(a, b, p1x, p1y, p2x, p2y, p3x, p3y, p4x, p4y, ratio, cc)
  return newPulleyJoint(a, b, p1y, p2x, p2y, p3x, p3y, p4x, p4y, ratio, cc)
end

local newGearJoint = lp.newGearJoint
function reg.World:newGearJoint(a, b, j1, j2, ratio, cc)
  return newGearJoint(j1, j2, ratio, cc)
end

local newWeldJoint = lp.newWeldJoint
function reg.World:newWeldJoint(a, b, p1x, p1y, cc)
  return newWeldJoint(a, b, p1x, p1y, cc)
end

local newFrictionJoint = lp.newFrictionJoint
function reg.World:newFrictionJoint(a, b, p1x, p1y, cc)
  return newFrictionJoint(a, b, p1x, p1y, cc)
end

--local newWheelJoint = lp.newWheelJoint
--function reg.World:newWheelJoint(a, b, p1x, p1y, ax, ay, cc)
--  return newWheelJoint(a, b, p1x, p1y, ax, ay, cc)
--end

local newWheelJoint = lp.newWheelJoint
function reg.World:newWheelJoint(body1, body2, x1, y1, x2, y2, ax, ay, collideConnected)
  return newWheelJoint(body1, body2, x1, y1, x2, y2, ax, ay, collideConnected)
end

local newMouseJoint = lp.newMouseJoint
function reg.World:newMouseJoint(a, x, y)
  local joint = newMouseJoint(a, x, y)
  return joint
end

--[[Contact lists
information about potentially colliding fixtures
each body in Box2D keeps a list of its potential contacts
iterate body contacts
local contacts = body:getContacs()
for i, contact in ipairs(contacts) do
iterate all contacts:
local contacts = world:getContacts()
Contact lists may contain potential collisions of fixtures that may not 
be touching at all. Using the "contact:IsTouching()" function tells us 
if there an actual collision.
Each contact point has an impulse associated with it. 
Box2D describes the magnitude of these impulses in two parts. 
The direction of these impulses is determined by the "collision normal".
Normal impulses (shown in yellow) push the two fixtures apart so that they are not inter-penetrating.
Tangent impulses are applied at 90-degrees relative to the "collision normal".
Tangent impulses are determined by the "friction" of fixtures and can cause the body to spin and roll.
Keep in mind that the exact point where an impulse is being applied is important too. 
The location of the contact point along with the center of mass of each body may produce 
torque and cause the body to spin.
--]]
function reg.Body:IsTouching(other)
  -- iterate contacts
  local contacts = self:getContacts()
  for i, contact in ipairs(contacts) do
    -- make sure there's actual contact
    if contact:isTouching() then
      --A contact between two fixtures may have 2, 1 or 0 contact points.
      -- look for a specific body
      local f1, f2 = contact:getFixtures()
      if f1:getBody() == other or f2:getBody() == other then
        --returns true even if there is no "solid" contact (sensors)
        return true
      end
    end
  end
  return false
end

--[[Is the body moving?
One approach is to compare the linear velocity of the body against a given 
threshold. This method returns true for a body resting on top of a moving platform.
--]]
function reg.Body:isMoving(treshold)
  treshold = treshold or 0
  local lvx, lvy = self:getLinearVelocity()
  return (lvx*lvx + lvy*lvy) > treshold*treshold
end

--[[Is the body rotating?
By looking at the angular velocity of a body we can determine if it's spinning. 
Note that the angular velocity could be negative. This approach would always 
return false for bodies with "fixed rotation".
--]]
function reg.Body:isRotating(treshold)
  treshold = treshold or 0
  local angular = self:getAngularVelocity()
  return angular < -treshold or angular > treshold
end

--[[Is the body moving along an axis?
This is a common question, when you want to find out if an object is "falling". 
We can use the dot product to check for movement along a given axis (ax, ay).
For example, you can use the gravity vector as an axis.
--]]
function reg.Body:isMovingOnAxis(ax, ay)
  local lvx, lvy = self:getLinearVelocity(lv)
  return ax*lvx + ay*lvy > 0
end

--[[Is the body pushed along an axis?
Using the contact information, we can check if a body is "pushed" by other 
bodies in a given direction. This applies for resting contact too. Suppose 
that we have a rectangle supported on top of a moving platform. Using the 
gravity axis, we check if the platform is "pushing up" the rectangle. This 
is one of the most accurate ways to tell if a given body is "on the ground".
--]]
function reg.Body:isPushedOnAxis(ax, ay)
  local contacts = self:getContacts()
  for i, contact in ipairs(contacts) do
    if contact:isTouching() then
      local x1, y1, x2, y2 = contact:getPositions()
      if x1 and y1 then
        local nx, ny = contact:getNormal()
        local f1, f2 = contact:getFixtures()
        local other = f1:getBody()
        if other ~= self then
          nx, ny = -nx, -ny
        end
        if ax*nx + ay*ny > 0 then
          return true
        end
      end
    end
  end
  return false
end

--[[Is the fixture flat on its side?
Consider an example of a rectangle resting on top of a platform. The number 
of points in each contact tells us how the body is being supported. If there 
is one point, the rectangle is supporting itself by a single vertex. If there 
are two points, the rectangle is flat on its side. There cannot be more than 
two contact points between a pair of (convex) fixtures.
Note that it's important to consider the direction of each collision normal 
in order to determine if a contact is pushing the body up, down, sideways or 
both! One fixture can be in contact with several others, so it's possible for 
the poor rectangle to be "sandwiched" or pushed in opposite directions at the 
same time.
--]]
--[[What is the "right" coefficient of restitution?
The restitution coefficient affects colliding bodies along the normal axis and is 
described as "the degree of relative kinetic energy retained after a collision". 
Box2D provides the following function to get the restitution between two colliding fixtures:
restitution = contact:getRestitution()
Depending on the restitution, we can categorize collisions in three types:
Perfectly elastic (restitution = 1; No kinetic energy is lost)
Example: perfectly elastic ball that can bounce forever
Elastic (restitution > 0 and restitution < 1)
Example: bouncing basketball
Inelastic (restitution = 0; the colliding objects remain together after the impact)
Example: ball made of soft clay that sticks to floor when dropped
As a general reference, let's look at the restitution coefficients of different types of balls:
0.858 golf ball
0.804 billiard ball
0.712 tennis ball
0.658 glass marble
0.597 steel ball bearing
--]]
--[[What is the relative velocity of any two bodies?
One simple approach is to subtract the linear velocities of the two bodies 
velocity1 = firstBody:getLinearVelocity()
velocity2 = secondBody:getLinearVelocity()
velocityDiff = velocity1 - velocity2
Next, we find how fast the two bodies are moving towards each other
 --direction vector
direction = firstBody:getPosition() - secondBody:getPosition()
directionNormal = normalize(direction)
 --relative speed (in Meters per second)
relativeSpeed = dotProduct(velocityDiff, directionNormal)
The resulting "relative speed" is:
positive when one or both bodies are moving towards each other,
zero when both bodies are immobile or moving together in parallel,
negative when one or both bodies are moving away from each other.
--]]
--[[What is the relative velocity at the moment of collision?
Collisions in Box2D are resolved when the "world:Step" function is called. 
That's why there is the "ContactListener" class - so our game is notified 
of collision events that happen during "world:Step".
The "ContactListener" class has four callbacks: "BeginContact", "EndContact", "PreSolve" and "PostSolve". 
Just remember that you really, really want to be using "Pre-" and "PostSolve" 
to buffer only the most essential collision data. Bodies and fixtures that 
are irrelevant to your game should be ignored altogether. Collision callbacks 
may be evoked several times per fixture which could easily amount to hundreds, 
even thousands of calls per frame.
The following method should work best during the "BeginContact" or "PreSolve" callbacks, 
before the velocity of the body is modified by Box2D. Given a contact point between two 
fixtures, Chris Campbell shows us a clever way to determine the "relative velocity of 
the contact points on each body". Campbell's approach is useful because it implicitly 
takes into account the bodies' angular velocity.
  local x1, y1, x2, y2 = contact:getPositions()
  if x1 and y1 then
    local f1, f2 = contact:getFixtures()
    local b1 = f1:getBody()
    local b2 = f2:getBody()
    -- Campbell's method
    local lvx1, lvy1 = b1:getLinearVelocityFromWorldPoint(x1, y1)
    local lvx2, lvy2 = b2:getLinearVelocityFromWorldPoint(x1, y1)
    -- velocity difference vector
    local dvx, dvy = lvx1 - lvx2, lvy1- lvy2
    -- impact speed (in Meters per second)
    local nx, ny = contact:getNormal()
    local impactSpeed = dvx*nx + dvy*ny -- dot product
--]]
